"""
Docker Wrapper

This module contains the `docker_wrapper` function for providing vantage6
algorithms with uniform input and output handling.
"""

import os
import pickle
import io
from abc import ABC, abstractmethod
import pandas

from vantage6.tools.dispatch_rpc import dispact_rpc
from vantage6.tools.util import info
from vantage6.tools import deserialization, serialization
from vantage6.tools.data_format import DataFormat
from vantage6.tools.exceptions import DeserializationException
from SPARQLWrapper import SPARQLWrapper, CSV
from typing import BinaryIO

_DATA_FORMAT_SEPARATOR = '.'
_MAX_FORMAT_STRING_LENGTH = 10

_SPARQL_RETURN_FORMAT = CSV


def docker_wrapper(module: str):
    wrapper = DockerWrapper()
    wrapper.wrap_algorithm(module)


def sparql_wrapper(module: str):
    wrapper = SparqlDockerWrapper()
    wrapper.wrap_algorithm(module)


class WrapperBase(ABC):

    def wrap_algorithm(self, module):
        """
            Wrap an algorithm module to provide input and output handling for the
            vantage6 infrastructure.

            Data is received in the form of files, whose location should be specified
            in the following environment variables:
            - `INPUT_FILE`: input arguments for the algorithm
            - `OUTPUT_FILE`: location where the results of the algorithm should be
              stored
            - `TOKEN_FILE`: access token for the vantage6 server REST api
            - `DATABASE_URI`: either a database endpoint or path to a csv file.

            The wrapper is able to parse a number of input file formats. The available
            formats can be found in `vantage6.tools.data_format.DataFormat`. When the
            input is not pickle (legacy), the format should be specified in the first
            bytes of the input file, followed by a '.'.

            It is also possible to specify the desired output format. This is done by
            including the parameter 'output_format' in the input parameters. Again, the
            list of possible output formats can be found in
            `vantage6.tools.data_format.DataFormat`.

            It is still possible that output serialization will fail even if the
            specified format is listed in the DataFormat enum. Algorithms can in
            principle return any python object, but not every serialization format will
            support arbitrary python objects. When dealing with unsupported algorithm
            output, the user should use 'pickle' as output format, which is the
            default.

            The other serialization formats support the following algorithm output:
            - built-in primitives (int, float, str, etc.)
            - built-in collections (list, dict, tuple, etc.)
            - pandas DataFrames

            :param module: module that contains the vantage6 algorithms
            :return:
            """
        info(f"wrapper for {module}")

        # read input from the mounted inputfile.
        input_file = os.environ["INPUT_FILE"]
        info(f"Reading input file {input_file}")

        input_data = load_input(input_file)

        # all containers receive a token, however this is usually only
        # used by the master method. But can be used by regular containers also
        # for example to find out the node_id.
        token_file = os.environ["TOKEN_FILE"]
        info(f"Reading token file '{token_file}'")
        with open(token_file) as fp:
            token = fp.read().strip()

        database_uri = os.environ["DATABASE_URI"]
        info(f"Using '{database_uri}' as database")
        # with open(data_file, "r") as fp:
        data = self.load_data(database_uri, input_data)

        # make the actual call to the method/function
        info("Dispatching ...")
        output = dispact_rpc(data, input_data, module, token)

        # write output from the method to mounted output file. Which will be
        # transfered back to the server by the node-instance.
        output_file = os.environ["OUTPUT_FILE"]
        info(f"Writing output to {output_file}")

        output_format = input_data.get('output_format', None)
        write_output(output_format, output, output_file)

    @staticmethod
    @abstractmethod
    def load_data(database_uri, input_data):
        pass


class DockerWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri, input_data):
        return pandas.read_csv(database_uri)


class SparqlDockerWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri, input_data):
        query = input_data['query']
        return SparqlDockerWrapper._query_triplestore(database_uri, query)

    @staticmethod
    def _query_triplestore(endpoint: str, query: str):
        sparql = SPARQLWrapper(endpoint, returnFormat=_SPARQL_RETURN_FORMAT)
        sparql.setQuery(query)

        result = sparql.query().convert().decode()

        return pandas.read_csv(io.StringIO(result))


def write_output(output_format, output, output_file):
    """
    Write output to output_file using the format from output_format.

    If output_format == None, write output as pickle without indicating format (legacy method)

    :param output_format:
    :param output:
    :param output_file:
    :return:
    """
    with open(output_file, 'wb') as fp:
        if output_format:
            # Indicate output format
            fp.write(output_format.encode() + b'.')

            # Write actual data
            output_format = DataFormat(output_format.lower())
            serialized = serialization.serialize(output, output_format)
            fp.write(serialized)
        else:
            # No output format specified, use legacy method
            fp.write(pickle.dumps(output))


def load_input(input_file):
    """
    Try to read the specified data format and deserialize the rest of the
    stream accordingly. If this fails, assume the data format is pickle.

    :param input_file:
    :return:
    """
    with open(input_file, "rb") as fp:
        try:
            input_data = _read_formatted(fp)
        except DeserializationException:
            info('No data format specified. '
                 'Assuming input data is pickle format')
            fp.seek(0)
            try:
                input_data = pickle.load(fp)
            except pickle.UnpicklingError:
                raise DeserializationException('Could not deserialize input')
    return input_data


def _read_formatted(file: BinaryIO):
    data_format = str.join('', list(_read_data_format(file)))
    data_format = DataFormat(data_format.lower())
    return deserialization.deserialize(file, data_format)


def _read_data_format(file: BinaryIO):
    """
    Try to read the prescribed data format. The data format should be specified
    as follows: DATA_FORMAT.ACTUAL_BYTES. This function will attempt to read
    the string before the period. It will fail if the file is not in the right
    format.

    :param file: Input file received from vantage infrastructure.
    :return:
    """
    success = False

    for i in range(_MAX_FORMAT_STRING_LENGTH):
        try:
            char = file.read(1).decode()
        except UnicodeDecodeError:
            # We aren't reading a unicode string
            raise DeserializationException('No data format specified')

        if char == _DATA_FORMAT_SEPARATOR:
            success = True
            break
        else:
            yield char

    if not success:
        # The file didn't have a format prepended
        raise DeserializationException('No data format specified')



import json
import pickle

import pandas as pd
from vantage6.tools.data_format import DataFormat
from vantage6.tools.util import info

_serializers = {}


def serialize(data, data_format: DataFormat):
    """
    Look up serializer for `data_format` and use this to serialize `data`.

    :param data:
    :param data_format:
    :return:
    """
    return _serializers[data_format](data)


def serializer(data_format: DataFormat):
    """
    Register function as serializer by adding it to the `_serializers` map with
    key `data_format`. This function should ideally support a multitude of
    python objects.

    There are two ways to extend serialization functionality:

    1. Create and register a new serialization function for a previously
       unsupported serialization format.
    2. Implement support for additional objects within an existing serializer
       function.

    :param data_format:
    :return:
    """

    def decorator_serializer(func):
        # Register serialization function
        _serializers[data_format] = func

        # Return function without modifications so it can also be run without
        # retrieving it from `_serializers`.
        return func

    return decorator_serializer


@serializer(DataFormat.JSON)
def serialize_to_json(data):
    info(f'Serializing type {type(data)} to json')

    if isinstance(data, pd.DataFrame) | isinstance(data, pd.Series):
        return _serialize_pandas(data)

    return _default_serialization(data)


def _default_serialization(data):
    info('Using default json serialization')
    return json.dumps(data).encode()


def _serialize_pandas(data):
    info('Running pandas json serialization')
    return data.to_json().encode()


@serializer(DataFormat.PICKLE)
def serialize_to_pickle(data):
    info('Serializing to pickle')
    return pickle.dumps(data)



import json
import pickle
from pathlib import Path
from unittest.mock import patch, MagicMock

import pandas as pd
from pytest import raises

from vantage6.tools import docker_wrapper
from vantage6.tools.exceptions import DeserializationException

MODULE_NAME = 'algorithm_module'
DATA = 'column1,column2\n1,2'
TOKEN = 'This is a fake token'
INPUT_PARAMETERS = {'method': 'hello_world'}
JSON_FORMAT = 'json'
SEPARATOR = '.'
SAMPLE_DB = pd.DataFrame([[1, 2]], columns=['column1', 'column2'])
PICKLE_FORMAT = 'pickle'

MOCK_SPARQL_ENDPOINT = 'sparql://some_triplestore'

def test_old_pickle_input_wrapper(tmp_path):
    """
    Testing if wrapper still parses legacy input.
    """
    input_file = tmp_path / 'input.pkl'

    with input_file.open('wb') as f:
        pickle.dump(INPUT_PARAMETERS, f)

    output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)
    assert file_echoes_db(output_file)


def test_json_input_without_format_raises_deserializationexception(tmp_path):
    """
    It should only be possible to provide json input if it is preceded by the string "json." in unicode. Otherwise a
    `DeserializationException` should be thrown.
    """
    input_file = tmp_path / 'input.json'

    with input_file.open('wb') as f:
        f.write(json.dumps(INPUT_PARAMETERS).encode())

    with raises(DeserializationException):
        run_docker_wrapper_with_echo_db(input_file, tmp_path)


def test_json_input_with_format_succeeds(tmp_path):
    input_file = tmp_path / 'input.txt'

    with input_file.open('wb') as f:
        f.write(f'JSON{SEPARATOR}'.encode())
        f.write(json.dumps(INPUT_PARAMETERS).encode())

    output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)
    assert file_echoes_db(output_file)


def test_pickle_input_with_format_succeeds(tmp_path):
    input_file = create_pickle_input(tmp_path)
    output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)
    assert file_echoes_db(output_file)


def test_wrapper_serializes_pickle_output(tmp_path):
    input_parameters = {'method': 'hello_world', 'output_format': PICKLE_FORMAT}
    input_file = create_pickle_input(tmp_path, input_parameters)

    output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)

    with output_file.open('rb') as f:
        # Check whether the output starts with `pickle.` to indicate the pickle data format
        assert f.read(len(PICKLE_FORMAT) + 1).decode() == f'{PICKLE_FORMAT}.'

        result = pickle.loads(f.read())
        pd.testing.assert_frame_equal(SAMPLE_DB, result)


def test_wrapper_serializes_json_output(tmp_path):
    input_parameters = {'method': 'hello_world', 'output_format': JSON_FORMAT}
    input_file = create_pickle_input(tmp_path, input_parameters)

    output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)

    with output_file.open('rb') as f:
        # Check whether the data is preceded by json format string
        assert f.read(len(JSON_FORMAT) + 1).decode() == f'{JSON_FORMAT}.'

        # Since the echo_db algorithm was triggered, output will be table that can be read by pandas.
        result = pd.read_json(f.read())
        pd.testing.assert_frame_equal(SAMPLE_DB, result)


def create_pickle_input(tmp_path, input_parameters=None):
    if input_parameters is None:
        input_parameters = INPUT_PARAMETERS

    input_file = tmp_path / 'input.pkl'
    with input_file.open('wb') as f:
        f.write(f'PICKLE{SEPARATOR}'.encode())
        f.write(pickle.dumps(input_parameters))
    return input_file


def file_echoes_db(output_file):
    with output_file.open('rb') as f:
        result = pickle.load(f)
        target = SAMPLE_DB

        return target.equals(result)


def run_docker_wrapper_with_echo_db(input_file, tmp_path):
    """
    Run the `echo_db` testing algorithm through the wrapper code. The wrapper communicates through files whose locations
    are stored in the `INPUT_FILE`, `TOKEN_FILE` and `DATABASE_URI` environment variables. The output of the
    algorithm is stored in the location specified in environment variable `OUTPUT_FILE`.

    :param input_file: input arguments to the wrapper and algorithm
    :param tmp_path: temporary path to store additional files.
    :return:
    """
    db_file = tmp_path / 'db_file.csv'
    token_file = tmp_path / 'token.txt'
    output_file = tmp_path / 'output_file.pkl'
    db_file.write_text(DATA)
    token_file.write_text(TOKEN)
    with patch('vantage6.tools.docker_wrapper.os') as mock_os:
        mock_os.environ = {
            'INPUT_FILE': input_file,
            'TOKEN_FILE': token_file,
            'OUTPUT_FILE': output_file,
            'DATABASE_URI': db_file
        }

        docker_wrapper.docker_wrapper(MODULE_NAME)
    return output_file


@patch('vantage6.tools.docker_wrapper.dispact_rpc')
@patch('vantage6.tools.docker_wrapper.os')
@patch('vantage6.tools.docker_wrapper.SPARQLWrapper')
def test_sparql_docker_wrapper_passes_dataframe(SPARQLWrapper: MagicMock, os: MagicMock, dispact_rpc: MagicMock,
                                        tmp_path: Path):
    input_file = tmp_path / 'input_file.pkl'
    token_file = tmp_path / 'token.txt'
    output_file = tmp_path / 'output.pkl'

    environ = {'INPUT_FILE': str(input_file),
               'TOKEN_FILE': str(token_file),
               'DATABASE_URI': MOCK_SPARQL_ENDPOINT,
               'OUTPUT_FILE': str(output_file)}

    os.environ = environ

    input_args = {'query': 'select *'}

    with input_file.open('wb') as f:
        pickle.dump(input_args, f)

    with token_file.open('w') as f:
        f.write(TOKEN)

    dispact_rpc.return_value = pd.DataFrame()
    SPARQLWrapper.return_value.query.return_value.convert.return_value = DATA.encode()

    docker_wrapper.sparql_wrapper(MODULE_NAME)

    dispact_rpc.assert_called_once()

    target_df = pd.DataFrame([[1, 2]], columns=['column1', 'column2'])
    pd.testing.assert_frame_equal(target_df, dispact_rpc.call_args[0][0])